/* $Id: entity.h,v 1.1.1.1 2002/07/07 14:03:34 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* Entities: readable, rewindable unicode objects */

/* #include <stdio.h> */

struct entity {
	struct entityop	*op;
	int refcnt;
	int inuse;		/* used by parser for recursion check */
	int flags;		/* used by parser */
#define ENTITY_ISEXTERNAL	0x1	/* for CRLF->LF translation */
};

/* method table */
struct entityop {
	void	(*rewind)(struct entity *);	/* open/rewind */
	wchar_t	(*getch)(struct entity *);
	int 	(*eof)(struct entity *);
	void 	(*free)(struct entity *);
	const char *(*name)(struct entity *);	/* user-readable name */
	void	(*setenc)(struct entity *, const char *);
};

#define ENTITY_REWIND(e)	(*(e)->op->rewind)(e)
#define ENTITY_EOF(e)		(*(e)->op->eof)(e)
#define ENTITY_GETCH(e)		(*(e)->op->getch)(e)
#define ENTITY_NAME(e)		(*(e)->op->name)(e)
#define ENTITY_SETENC(e, enc)	(*(e)->op->setenc)(e, enc)

#define ENTITY_INCREF(e)	((e)->refcnt++)
#define ENTITY_DECREF(e)			\
    do {					\
	(e)->refcnt--;				\
	if ((e)->refcnt == 0)			\
		(*(e)->op->free)(e);		\
    } while (0)

/* function version of the macro */
void	entity_decref(struct entity *);

/* convenience macro for constructor functions */
#define ENTITY_INIT(e, eo)			\
    do {					\
	(e)->refcnt = 1;			\
	(e)->inuse = 0;				\
	(e)->op = (eo);				\
	(e)->flags = 0;				\
    } while (0)

/* entity creation normally malloc/copy the data, unless nocopy is set */
struct	entity *entity_mem(const wchar_t *start, const wchar_t *end, 
				int nocopy);
struct	entity *entity_utf8mem(const char *start, const char *end, int nocopy);
struct	entity *entity_eof(void);
struct	entity *entity_file(FILE *, const char *);
