#!perl
# $Id: dtd2grammar.pl,v 1.2 2002/07/18 23:18:12 d Exp $
# David Leonard, 2002. Public domain.

#-- Convert a DTD file into a collection of Python tables to be
#   used by HTMLExpander.py

#$debug = 1;
%entity = ();
%element = ();

#-- print the C source prologue

print "

/* Generated by $0 */

/* Metasymbols */
static char HTMLVersion[] = \"".$entity{"HTML.Version"}."\",
	PCData[] = \"#PCData\",
	O[] = \"Optional\",
	R[] = \"Required\";

/* Elements */
";

$ungetbuf = "";
sub getline {
	my $ret;
	if ($ungetbuf ne "") {
		if ($ungetbuf =~ s/^[^>]*>//) { 
			$ret = $&;
		} else {
			$ret = $ungetbuf;
			$ungetbuf = "";
		}
	} else { 
		$ret = <STDIN>;
	}
	#print STDERR "READ: $ret\n";
	return $ret;
}

sub unget {
	my $d = shift;
	$ungetbuf = $d . $ungetbuf;
}

sub elvar {
	my $en = shift;
	if (!defined($element{$en})) {
		my $var = $en;
		$var =~ s/[^0-9a-z_]/_/gi;
		$element{$en} = $var;
		print "char ".$var."[] = \"".lc($en)."\";\n";
	}
	return $element{$en};
}

#-- read lines from the DTD file, ending in '>'
$/ = ">";
while ($_ = &getline()) {
	
	#-- zap comments
	while (m/<!--/) {
		s/<!--(.|\n)*?-->//g;
		if (m/<!--/) {
			$_ .= &getline();
		}
	}

	#-- expand all entities now
	if (s/%([a-zA-Z][a-zA-Z0-9.]*);?/
		if (!defined($entity{$1})) { 
			print STDERR "Unknown entity %$1;\n";
			"???";
		} else {
			$entity{$1};
		}
	/eg) {
		unget($_);
		next;
	}

	#-- strip off any leading text before the opening '<'
	s/^[^<]*//;
	$text = $&;

	#print STDERR "COMMAND: $_\n";

	#-- handle "character data" in the DTD
	if ($text !~ m/^\s*$/) {
		#print "CDATA: $text\n";
	}

	#-- handle CDATA
	if (s/^<!\[\s*(\S+)?\s*\[//) {
		my $expand = $1;
		while (!s/\]\]>(.*)//) { $_ .= &getline(); }
		&unget($1);
		if (uc($expand) ne "IGNORE") { &unget($_); }
		next;
	}

	$origline = $_;

	#-- get rid of comments of the form <! ... -- comment -- ... >
	s/\s+/ /g;
	s/--([^-]|-[^-])*--//g;
	while (m/--/) {
		my $nl = &getline();
		$_ .= $nl;
		s/\s+/ /g;
		s/--([^-]|-[^-])*--//g;
	}

	#-- remove the trailing '>' and what follows
	s/\s*>(.*)//;
	&unget($1);

	#-- handle new entity definitions
	if (m/<!ENTITY % (\S+)( PUBLIC)? "([^"]*)"\s*("[^"]*")?/) {
		my $name = $1;
		my $public = $2;
		my $value = $3;
		my $filename = $4;
		if (defined($entity{$name})) {  
			print STDERR "$.: duplicate entity %$name;\n";
		} elsif ($public eq " PUBLIC") {
			if ($value =~ m,-//W3C//DTD HTML 4.01 Transitional//EN,) { $filename = "loose.dtd"; }
			$filename =~ s/^"(.*)"$/$1/;
			print STDERR "reading file $filename into %$name;\n";
			open(I, "<$filename")
				|| die "$filename: $!";
			$value = "";
			while (<I>) { $value .= $_; }
			close(I);
			$entity{$name} = $value;
		} else {
			#print STDERR "%$name; = '$value'\n";
			$entity{$name} = $value;
		}
		next;
	}
	elsif (m/<!ENTITY/) {
		#print "missed: $_\n";
	}

	#-- handle an element definition
	if (m/^<!ELEMENT ([^(\s]\S*|\([^)]*\)) ([-O]) ([-O]) (.*?)(-\(([^)]+)\))?$/) {
		my $element = $1;
		my $open = $2;
		my $close = $3;
		my $contain = $4;
		my $exclude = $6;
		$contain =~ s/\s+//g;
		$exclude =~ s/\s+//g;
		$element =~ s/^\(\s*(.*?)\s*\)$/$1/;
		my @e = split(/\s*\|\s*/, $element);

		$contain = "" if $contain eq "EMPTY";

		my %ins = ();
		for $i (split(/[|(),+*?&]+/, $contain)) { $ins{$i} = 1; }
		for $i (split(/[|(),]+/, $exclude)) { $ins{$i} = 0; }
		$ins{''} = 0;
		my @inside = grep($ins{$_}==1, sort keys %ins);
		$inside = join(' ', @inside);
		$open = "R"  if $open eq "-";
		$close = "R" if $close eq "-";

		for (@inside) {
			if ($_ eq "#PCDATA" or $_ eq "CDATA") { $_ = "PCData"; }
			else { $_ = &elvar($_); }
		}

		foreach $e (@e) {
			#-- print some debugging about the element
			print STDERR "<$e>:\n" .
			      "\topen:    $open\n".
			      "\tclose:   $close\n".
			      "\tcontain: $contain\n".
			      "\texclude: $exclude\n".
			      "\tinside:  $inside\n".
			      "\n" if $debug;

			#-- add to the 'M' definition
			if ($M) { $M .= ",\n" }
			$M .= 
			   " { ".&elvar($e).", $open, $close, { ".
				join(', ', (@inside, "(char *)0")).
				" } }";
		}
	}

	#-- process an attribute definition
	if (m/^<!ATTLIST ([^(\s]\S*|\([^)]*\)) (.*)/) {
		my $element = $1;
		my $a = $2;
		$a =~ s/\(/ (/g;
		$a =~ s/\)/) /g;
		$a =~ s/\s+/ /g;
		$a =~ s/^\s*//;
		$a =~ s/\s*$//;
		$a =~ s/\s*([,|])\s*/$1/g;
		$a =~ s/\(\s*/(/g;
		$a =~ s/\s*\)/)/g;
		my %keep = ();
		my $keepid = 0;

		#-- split the element name - which maybe a list
		$element =~ s/^\(\s*(.*?)\s*\)$/$1/;
		my @e = split(/\s*\|\s*/, $element);

		#-- pull out things with spaces in them for now
		$a =~ s/(#FIXED )?'([^']*)'/ 
			$keep{$keepid} = $&; "%".($keepid++).";" /eg;
		$a =~ s/(#FIXED )?"([^"]*)"/
			$keep{$keepid} = $&; "%".($keepid++).";" /eg;

		#-- split the contents on spaces
		my @a = split(/\s+/, $a);

		for (@a) {
			s/%(.*);/$keep{$1}/eg;
		}

		#-- must be a triple of values
		if (($#a + 1) % 3 != 0) {
			die "not a multiple of 3 args: $element(".
				join(',',@a).")\n$origline";
		}

	      for $e (@e) {
		for ($i = 0; $i < $#a; $i += 3) {
		    my $name = uc($a[$i]);
		    my $type = $a[1+$i];
		    my $value = $a[2+$i];

		    print STDERR "${e}.${name}: ${type} = ${value}\n"
			    if $debug;

		    my $t, $v;
		    if ($type eq "CDATA") { $t = "PCData"; }
		    elsif ($type eq "NAME") { $t = "Name"; }
		    elsif ($type eq "ID") { $t = "Id"; }
		    elsif ($type eq "IDREF") { $t = "IdRef"; }
		    elsif ($type eq "IDREFS") { $t = "IdRefs"; }
		    elsif ($type eq "NUMBER") { $t = "Number"; }
		    elsif ($type =~ m/^\(/) {
			if ($A) { $A .= ",\n" }
			$A .= "\t{ ".&elvar($e).", ".&elvar($name)." }";
		    }
		    else {
			die "Unknown type $type";
		    }
	        }
	      }
	}
}

&elvar("HTML");

print "

/* Tag requirements and children */
struct M {
	char *tag, *open, *close;
	char *content[100];
} Mtab[] = {
$M
};
static int nMtab = sizeof Mtab / sizeof Mtab[0];

/* Attributes that have an enumeration type */
struct A { char *tag, *attr; } Atab[] = {
$A
};
static int nAtab = sizeof Atab / sizeof Atab[0];
";
