/* $Id: intern.c,v 1.3 2002/07/16 06:36:42 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* Internalizing of strings */

#include <stdlib.h>
#include "utf8.h"
#include "xhash.h"
#include "vector.h"
#include "wstr.h"
#include "intern.h"

struct xhash *intern_hash = NULL;

static void
intern_init()
{
	intern_hash = xhash_new(NULL, (void (*)(void *))wstr_free);
}

/* The hash takes ownership of w */
wstr_t
intern_wstr(w)
	wstr_t w;
{
	wstr_t iw;

	if (w == NULL)
		return NULL;
	if (intern_hash == NULL)
		intern_init();
	iw = (wstr_t)xhash_lookup(intern_hash, WSTR_START(w), WSTR_SIZE(w));
	if (iw == NULL) {
		xhash_store(intern_hash, WSTR_START(w), WSTR_SIZE(w), w);
		iw = w;
	} 
	if (iw != w)
		wstr_free(w);
	return iw;
}

/* The hash does not take ownership of w */
wstr_t
intern_wstr_noown(w)
	wstr_t w;
{
	wstr_t iw;
	size_t sz;
	wchar_t *s;

	if (w == NULL)
		return NULL;
	if (intern_hash == NULL)
		intern_init();
	sz = WSTR_SIZE(w);
	s = WSTR_START(w);
	iw = (wstr_t)xhash_lookup(intern_hash, s, sz);
	if (iw == NULL) {
		iw = wstr_dup(w);
		xhash_store(intern_hash, s, sz, iw);
	} 
	return iw;
}

/* Expand UTF8 to a wchar_t array, and then intern */
wstr_t
intern_utf8(s)
	const char *s;
{
	wchar_t *buf, *end, *p;
	int len;

	if (s == NULL)
		return NULL;
	if (intern_hash == NULL)
		intern_init();
	len = strlen(s);
	buf = (wchar_t *)alloca(len * sizeof (wchar_t));
	end = buf + len;
	for (p = buf; p < end && *s; p++)
		*p = utf8_get(&s);
	return intern(buf, (p - buf) * sizeof (wchar_t));
}

/* Intern the string */
wstr_t
intern(s, sz)
	const wchar_t *s;
	size_t sz;
{
	wstr_t iw;
	int i;

	if (intern_hash == NULL)
		intern_init();
	iw = (wstr_t)xhash_lookup(intern_hash, s, sz);
	if (iw == NULL) {
		iw = wstr_new();
		for (i = 0; i < sz / sizeof (wchar_t); i++)
			WSTR_ADD(iw, s[i]);
		xhash_store(intern_hash, s, sz, iw);
	}
	return iw;
}
