/* $Id: vector.c,v 1.3 2002/07/16 06:36:42 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* Vectors are auto-allocated arrays */

#include <stdio.h>
#include <stdlib.h>
#include "vector.h"

/* incrementally increase the size of a vector's allocation */
int
vector_grow(v, sz)
	vector_t *v;
	int sz;
{
	int newalloc = v->alloc;
	char *newstore;

	if (newalloc == 0) {
		newalloc = 256;
		newstore = malloc(newalloc * sz);
	} else {
		newalloc *= 2;
		newstore = realloc((char *)v->store, newalloc * sz);
	}
	if (newstore == NULL) {
		return 0;
	}
	v->store = newstore;
	v->alloc = newalloc;
	return 1;
}

void
vector_release(v)
	vector_t *v;
{
	if (v->alloc)
		free(v->store);
	v->alloc = 0;
	v->len = 0;
}

void
vector_copy(dst, src, sz)
	vector_t *dst, *src;
	int sz;
{
	dst->alloc = dst->len = src->len;
	if (dst->alloc) {
		dst->store = malloc(dst->alloc * sz);
		memcpy(dst->store, src->store, src->len * sz);
	} else
		dst->store = NULL;
}

void
vector_concat(dst, src, sz)
	vector_t *dst, *src;
	int sz;
{
	if (src->len) {
		while (dst->alloc < dst->len + src->len)
			vector_grow(dst, sz);
		memcpy(dst->store + (dst->len * sz), src->store, src->len * sz);
		dst->len += src->len;
	}
}
