/* $Id: vector.h,v 1.4 2002/07/17 03:07:28 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * Vectors are arrays that automatically grow.
 */

/* Hide library-internal symbols by prefixing */
#define vector_grow	dxml_vector_grow
#define vector_release	dxml_vector_release
#define vector_copy	dxml_vector_copy
#define vector_concat	dxml_vector_concat

#define VECTOR_DECL(t)							\
	struct {							\
		t *store;						\
		int alloc, len;						\
	}

typedef VECTOR_DECL(char) vector_t;	/* generic vector */
int	vector_grow(vector_t *, int);
void	vector_release(vector_t *);

#define VECTOR_LEN(v)		((v).len)
#define VECTOR_PTR(v)		((void *)(v).store)
#define VECTOR_SIZE(v)		((v).len * VECTOR_ESZ(v))
#define VECTOR_ESZ(v)		(sizeof *(v).store)

#define VECTOR_INIT(v)							\
	do {								\
		(v).alloc = 0;						\
		(v).len = 0;						\
	} while (0)

#define VECTOR_STATIC_INIT	{ 0, 0, 0 }

#define VECTOR_APPEND(v, d)						\
	do {								\
		if ((v).len < (v).alloc || 				\
		    vector_grow((vector_t *)&(v), VECTOR_ESZ(v)))	\
			(v).store[(v).len++] = (d);			\
	} while (0)

#define VECTOR_FREE(v)							\
	do {								\
		if ((v).alloc)						\
			vector_release((vector_t *)&(v));		\
	} while (0)

#define VECTOR_AT(v, i)		((v).store[i])

void	vector_copy(vector_t *, vector_t *, int);
void	vector_concat(vector_t *, vector_t *, int);

#define VECTOR_COPY(dst, src)	\
	vector_copy((vector_t *)&(dst), (vector_t *)&(src), \
		VECTOR_ESZ(src))

#define VECTOR_CONCAT(dst, src)	\
	vector_concat((vector_t *)&(dst), (vector_t *)&(src), \
		VECTOR_ESZ(src))

#define VECTOR_EQ(a, b)	\
	(VECTOR_LEN(a) == VECTOR_LEN(b) && \
	 VECTOR_ESZ(a) == VECTOR_ESZ(b) && \
	 memcmp(VECTOR_PTR(a), VECTOR_PTR(b), VECTOR_SIZE(a)) == 0)

#define VECTOR_SORT(v, cmp) \
	qsort(VECTOR_PTR(v), VECTOR_LEN(v), VECTOR_ESZ(v), cmp)
#define VECTOR_SEARCH(key, v, cmp) \
	bsearch(&(key), VECTOR_PTR(v), VECTOR_LEN(v), VECTOR_ESZ(v), cmp)

