/* $Id: wstr.c,v 1.2 2002/07/16 06:36:42 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* wide character vector/string */

#include <stdio.h>
#include <stdlib.h>

#include "utf8.h"
#include "vector.h"
#include "entity.h"
#include "wstr.h"

static void	wstre_rewind(struct entity *e);
static wchar_t	wstre_getch(struct entity *e);
static int	wstre_eof(struct entity *e);
static void	wstre_free(struct entity *e);
static const char *wstre_name(struct entity *e);

wstr_t
wstr_new()
{
	wstr_t w;

	w = (wstr_t)malloc(sizeof *w);
	if (w)
		VECTOR_INIT(*w);
	return w;
}

void
wstr_free(w)
	wstr_t w;
{
	if (w) {
		VECTOR_FREE(*w);
		free(w);
	}
}

const char *
wstr_str(w)
	wstr_t w;
{
	static char buf[80];
	int i;

	if (!w)
		return "(null)";
	for (i = 0; i < VECTOR_LEN(*w) && i < sizeof buf - 1; i++)
		if (VECTOR_AT(*w, i) < 0x80)
			buf[i] = VECTOR_AT(*w, i);
		else
			buf[i] = '?';
	buf[i] = '\0';
	return buf;
}

void
wstr_addwstr(w, x)
	wstr_t w, x;
{
	int i;

	if (!x)
		return;
	for (i = 0; i < VECTOR_LEN(*x); i++)
		VECTOR_APPEND(*w, VECTOR_AT(*x, i));
}

void
wstr_addentity(w, e)
	wstr_t w;
	struct entity *e;
{
	if (!e)
		return;
	if (e->inuse)
		return;
	ENTITY_REWIND(e);
	while (!ENTITY_EOF(e))
		VECTOR_APPEND(*w, ENTITY_GETCH(e));
}

int
wstr_eq(w, x)
	wstr_t w, x;
{
	int i;

	if (w == NULL || x == NULL)
		return 0;
	if (VECTOR_LEN(*w) != VECTOR_LEN(*x))
		return 0;
	for (i = 0; i < VECTOR_LEN(*w); i++)
		if (VECTOR_AT(*w, i) != VECTOR_AT(*x, i))
			return 0;
	return 1;
}

wstr_t
wstr_dup(w)
	wstr_t w;
{
	wstr_t nw;

	nw = wstr_new();
	wstr_addwstr(nw, w);
	return nw;
}

/*------------------------------------------------------------
 * a wstr entity
 */

static struct entityop wstr_ops = {
	wstre_rewind,
	wstre_getch,
	wstre_eof,
	wstre_free,
	wstre_name
};

struct wstr_entity {
	struct entity e;
	wstr_t	w;
	int	pos;
};

static void
wstre_rewind(e)
	struct entity *e;
{
	struct wstr_entity *we = (struct wstr_entity *)e;

	we->pos = 0;
}

static wchar_t
wstre_getch(e)
	struct entity *e;
{
	struct wstr_entity *we = (struct wstr_entity *)e;
	wchar_t c;

	if (we->pos >= VECTOR_LEN(*we->w))
		return -1;
	c = VECTOR_AT(*we->w, we->pos);
	we->pos++;
	return c;
}

static int
wstre_eof(e)
	struct entity *e;
{
	struct wstr_entity *we = (struct wstr_entity *)e;

	return we->pos >= VECTOR_LEN(*we->w);
}

static void
wstre_free(e)
	struct entity *e;
{
	struct wstr_entity *we = (struct wstr_entity *)e;

	wstr_free(we->w);
	free(we);
}

static const char *
wstre_name(e)
	struct entity *e;
{
	return NULL;
}

struct entity *
entity_wstr(w)
	wstr_t w;
{
	struct wstr_entity *we;

	we = (struct wstr_entity *)malloc(sizeof (struct wstr_entity));
	if (!we)
		return NULL;
	ENTITY_INIT(&we->e, &wstr_ops);
	we->w = w;
	we->pos = 0;
	return (struct entity *)we;
}

wstr_t
wstr_new_wchar(str, sz)
	wchar_t *str;
	size_t sz;
{
	wstr_t w = wstr_new();
	if (w) {
		while (sz > 0) {
			VECTOR_APPEND(*w, *str);
			str++;
			sz -= sizeof (wchar_t);
		}
	}
	return w;
}
