/* $Id: xfsm.h,v 1.2 2002/07/16 06:36:43 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* Finite state machines */

/* Hide library-internal names by prefixing with dxml */
#define xfsm_init		dxml_xfsm_init
#define xfsm_empty		dxml_xfsm_empty
#define xfsm_simple		dxml_xfsm_simple
#define xfsm_seq		dxml_xfsm_seq
#define xfsm_alt		dxml_xfsm_alt
#define xfsm_kleene		dxml_xfsm_kleene
#define xfsm_plus		dxml_xfsm_plus
#define xfsm_opt		dxml_xfsm_opt
#define xfsm_addtrans		dxml_xfsm_addtrans
#define xfsm_reduce		dxml_xfsm_reduce
#define xfsm_getstart		dxml_xfsm_getstart
#define xfsm_getnext		dxml_xfsm_getnext
#define xfsm_getpossible	dxml_xfsm_getpossible
#define xfsm_isaccept		dxml_xfsm_isaccept
#define xfsm_free		dxml_xfsm_free

struct xfsm;

/* Initialise a head pointer into a FSM graph */
void	xfsm_init(struct xfsm **h);

/*
 * Construction of an (unreduced) FSM from a regular language
 *	xfs_empty:	()
 *	xfs_simple:	s
 *	xfs_seq:	xy
 *	xfs_alt:	(x|y)
 *	xfs_kleene:	x*
 *	xfs_plus:	x+
 *	xfs_opt:	x?		-- one less state than (x|)
 * All functions return 1 on success or 0 on memory error.
 */

int	xfsm_empty(struct xfsm **h, struct xfsm **a, struct xfsm **b);
int	xfsm_simple(struct xfsm **h, const void *sym,
			struct xfsm **a, struct xfsm **b);
int	xfsm_seq(struct xfsm **h, struct xfsm *xs, struct xfsm *xf,
			struct xfsm *ys, struct xfsm *yf,
			struct xfsm **a, struct xfsm **b);
int	xfsm_alt(struct xfsm **h, struct xfsm *xs, struct xfsm *xf,
			struct xfsm *ys, struct xfsm *yf,
			struct xfsm **a, struct xfsm **b);
int	xfsm_kleene(struct xfsm **h, struct xfsm *xs, struct xfsm *xf,
			struct xfsm **a, struct xfsm **b);
int	xfsm_plus(struct xfsm **h, struct xfsm *xs, struct xfsm *xf,
			struct xfsm **a, struct xfsm **b);
int	xfsm_opt(struct xfsm **h, struct xfsm *xs, struct xfsm *xf,
			struct xfsm **a, struct xfsm **b);

/* If manual addition of transitions is required... (eek!) */
void	xfsm_addtrans(struct xfsm *a, const void *sym, struct xfsm *b);

/* Reduce the FSM to a deterministic (but sub-optimal) form. */
void	xfsm_reduce(struct xfsm **h);

/*
 * These next functions should only be used after calling xfsm_reduce() once.
 * Use xfsm_getstart() to find the start state, and transit to
 * other nodes by using xfsm_getnext(): invalid transitions return
 * NULL. A list of expected symbols at a state can be obtained by calling 
 * xfsm_getpossible(), iterating the index from 0 upwards until 
 * it returns 0. When finishing up, the state must be checked to be
 * a valid accept state with xfsm_isaccept().
 */
struct xfsm *xfsm_getstart(struct xfsm **h);
struct xfsm *xfsm_getnext(struct xfsm *, const void *symbol);
int	     xfsm_getpossible(struct xfsm *, int index, const void **symp);
int	     xfsm_isaccept(struct xfsm *);

void	xfsm_free(struct xfsm **h);


