/* $Id: xmlns.c,v 1.2 2002/07/16 06:36:42 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * Namespaces
 *
 * A namespace is a tree of nested scopes.
 * Each bound name really introduces a new scope.
 * Lookups are performed by walking towards the root of the
 * tree looking for a match.
 * Names and URIs are interned for speed.
 */

#include <stdlib.h>
#include "utf8.h"
#include "vector.h"
#include "wstr.h"
#include "intern.h"
#include "xmlns.h"
#include "xmlnsP.h"

struct xmlns {
	struct xmlns *scope;
	wstr_t	name;
	wstr_t	uri;
};

wstr_t xmlns_unknown = (wstr_t)-1;

struct xmlns *
xmlns_bind(ns, name, namesize, uri, urisize)
	struct xmlns *ns;
	const wchar_t *name;
	size_t namesize;
	const wchar_t *uri;
	size_t urisize;
{
	return xmlns_bindw(ns, intern(name, namesize),
		intern(uri, urisize));
}

struct xmlns *
xmlns_bind_utf8(ns, name, uri)
	struct xmlns *ns;
	const char *name;
	const char *uri;
{
	return xmlns_bindw(ns, intern_utf8(name), intern_utf8(uri));
}

struct xmlns *
xmlns_bindw(ns, name, uri)
	struct xmlns *ns;
	wstr_t name;
	wstr_t uri;
{
	struct xmlns *newns;

	newns = (struct xmlns *)malloc(sizeof (struct xmlns));
	if (newns) {
		newns->scope = ns;
		newns->name = name;
		newns->uri = uri;
	}
	return newns;
}

struct xmlns *
xmlns_free(ns)
	struct xmlns *ns;
{
	struct xmlns *scope;

	if (ns) {
		scope = ns->scope;
		free(ns);
	} else
		scope = NULL;
	return scope;
}

wstr_t
xmlns_lookup(ns, name)
	struct xmlns *ns;
	wstr_t name;
{
	for (; ns; ns = ns->scope)
		if (ns->name == name)
			return ns->uri;
	return xmlns_unknown;
}

/*
 * find a namespace's name, when given a URI.
 * (useful when constructing error messages)
 */
wstr_t
xmlns_finduri(ns, uri)
	struct xmlns *ns;
	wstr_t uri;
{
	for (; ns; ns = ns->scope)
		if (ns->uri == uri)
			return ns->name;
	return NULL;
}
