/* $Id: xpath.h,v 1.8 2002/07/17 01:04:42 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* XPath filters */

struct entity;
struct xpath;
struct xmlns;

typedef struct xpath *xpath_t;

xpath_t	 xpath_new(const wchar_t *, size_t size);
xpath_t	 xpath_new_utf8(const char *);		/* string is UTF-8 encoded */
xpath_t	 xpath_newe(struct entity *);
void	 xpath_free(xpath_t);

xml_t	*xpath_select(xpath_t, xml_t, struct xmlns *);
			/* null-terminated array; caller must free */
int	 xpath_eval_boolean(xpath_t, xml_t, struct xmlns *);
			/* -1 on error, else 0=false, 1=true */
char	*xpath_eval_utf8string(xpath_t, xml_t, struct xmlns *);
			/* UTF-8, nul-terminated; caller must free */
double	 xpath_eval_number(xpath_t, xml_t, struct xmlns *);
			/* xpath_last_error will not be NULL on error */

union xpath_value {
	xml_t	*x_nodeset;
	int	x_boolean;
	double	x_number;
	struct	{
		char	*data;
		size_t 	 size;
	} x_string;
};

int	 xpath_eval(xpath_t, xml_t, struct xmlns *, union xpath_value *);
#define XPATH_VALUE_ERROR	-1
#define XPATH_VALUE_NODESET	0
#define XPATH_VALUE_BOOLEAN	1
#define XPATH_VALUE_NUMBER	2
#define XPATH_VALUE_STRING_UTF8	3

/* bind $uri:name names */
void	 xpath_bind_utf8_nodeset(xpath_t, const char *, xml_t *, int);
void	 xpath_bind_utf8_boolean(xpath_t, const char *, int);
void	 xpath_bind_utf8_number(xpath_t, const char *, double);
void	 xpath_bind_utf8_string(xpath_t, const char *, const char *);
void	 xpath_unbind_utf8(xpath_t, const char *);

/* wchar_t variants */
void	 xpath_bind_nodeset(xpath_t path,
			const wchar_t *uri, size_t urisize,
			const wchar_t *name, size_t namesize,
			xml_t *nodes, int num_nodes);
void	 xpath_bind_boolean(xpath_t path,
			const wchar_t *uri, size_t urisize,
			const wchar_t *name, size_t namesize,
			int truth);
void	 xpath_bind_number(xpath_t path,
			const wchar_t *uri, size_t urisize,
			const wchar_t *name, size_t namesize,
			double number);
void	 xpath_bind_string(xpath_t,
			const wchar_t *uri, size_t urisize,
			const wchar_t *name, size_t namesize,
			const wchar_t *str, size_t strsize);
void	 xpath_unbind(xpath_t,
			const wchar_t *uri, size_t urisize,
			const wchar_t *name, size_t namesize);

extern const char *xpath_last_error;

int	xpath_puts(xpath_t, int (*)(const char *, void *), void *);
